@@ -10,11 +10,11 @@
  * the rights to use, copy, modify, merge, publish, distribute, sublicense,
  * and/or sell copies of the Software, and to permit persons to whom the
  * Software is furnished to do so, subject to the following conditions:
- * 
+ *
  * The above copyright notice and this permission notice (including the next
  * paragraph) shall be included in all copies or substantial portions of the
  * Software.
- * 
+ *
  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
@@ -22,7 +22,7 @@
  * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
  * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  * DEALINGS IN THE SOFTWARE.
- * 
+ *
  * Authors: Rickard E. (Rik) Faith <faith@valinux.com>
  */
 /*
@@ -57,6 +57,7 @@
 #endif
 
 #include <X11/X.h>
+#include <xserver_poll.h>
 #include "xf86.h"
 #include "xf86Priv.h"
 #include "xf86_OSlib.h"
@@ -83,8 +84,36 @@ typedef struct _xf86SigIOFunc {
 
 static Xf86SigIOFunc xf86SigIOFuncs[MAX_FUNCS];
 static int xf86SigIOMax;
-static int xf86SigIOMaxFd;
-static fd_set xf86SigIOMask;
+static struct pollfd *xf86SigIOFds;
+static int xf86SigIONum;
+
+static Bool
+xf86SigIOAdd(int fd)
+{
+    struct pollfd *n;
+
+    n = realloc(xf86SigIOFds, (xf86SigIONum + 1) * sizeof (struct pollfd));
+    if (!n)
+        return FALSE;
+
+    n[xf86SigIONum].fd = fd;
+    n[xf86SigIONum].events = POLLIN;
+    xf86SigIONum++;
+    xf86SigIOFds = n;
+    return TRUE;
+}
+
+static void
+xf86SigIORemove(int fd)
+{
+    int i;
+    for (i = 0; i < xf86SigIONum; i++)
+        if (xf86SigIOFds[i].fd == fd) {
+            memmove(&xf86SigIOFds[i], &xf86SigIOFds[i+1], (xf86SigIONum - i - 1) * sizeof (struct pollfd));
+            xf86SigIONum--;
+            break;
+        }
+}
 
 /*
  * SIGIO gives no way of discovering which fd signalled, select
@@ -93,24 +122,22 @@ static fd_set xf86SigIOMask;
 static void
 xf86SIGIO(int sig)
 {
-    int i;
-    fd_set ready;
-    struct timeval to;
+    int i, f;
     int save_errno = errno;     /* do not clobber the global errno */
     int r;
 
     inSignalContext = TRUE;
 
-    ready = xf86SigIOMask;
-    to.tv_sec = 0;
-    to.tv_usec = 0;
-    SYSCALL(r = select(xf86SigIOMaxFd, &ready, 0, 0, &to));
-    for (i = 0; r > 0 && i < xf86SigIOMax; i++)
-        if (xf86SigIOFuncs[i].f && FD_ISSET(xf86SigIOFuncs[i].fd, &ready)) {
-            (*xf86SigIOFuncs[i].f) (xf86SigIOFuncs[i].fd,
-                                    xf86SigIOFuncs[i].closure);
+    SYSCALL(r = xserver_poll(xf86SigIOFds, xf86SigIONum, 0));
+    for (f = 0; r > 0 && f < xf86SigIONum; f++) {
+        if (xf86SigIOFds[f].revents & POLLIN) {
+            for (i = 0; i < xf86SigIOMax; i++)
+                if (xf86SigIOFuncs[i].f && xf86SigIOFuncs[i].fd == xf86SigIOFds[f].fd)
+                    (*xf86SigIOFuncs[i].f) (xf86SigIOFuncs[i].fd,
+                                            xf86SigIOFuncs[i].closure);
             r--;
         }
+    }
     if (r > 0) {
         xf86Msg(X_ERROR, "SIGIO %d descriptors not handled\n", r);
     }
@@ -130,6 +157,26 @@ xf86IsPipe(int fd)
     return S_ISFIFO(buf.st_mode);
 }
 
+static void
+block_sigio(void)
+{
+    sigset_t set;
+
+    sigemptyset(&set);
+    sigaddset(&set, SIGIO);
+    xthread_sigmask(SIG_BLOCK, &set, NULL);
+}
+
+static void
+release_sigio(void)
+{
+    sigset_t set;
+
+    sigemptyset(&set);
+    sigaddset(&set, SIGIO);
+    xthread_sigmask(SIG_UNBLOCK, &set, NULL);
+}
+
 int
 xf86InstallSIGIOHandler(int fd, void (*f) (int, void *), void *closure)
 {
@@ -145,7 +192,7 @@ xf86InstallSIGIOHandler(int fd, void (*f) (int, void *), void *closure)
         if (!xf86SigIOFuncs[i].f) {
             if (xf86IsPipe(fd))
                 return 0;
-            OsBlockSIGIO();
+            block_sigio();
 #ifdef O_ASYNC
             if (fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_ASYNC) == -1) {
                 xf86Msg(X_WARNING, "fcntl(%d, O_ASYNC): %s\n",
@@ -173,12 +220,12 @@ xf86InstallSIGIOHandler(int fd, void (*f) (int, void *), void *closure)
             }
 #endif
             if (!installed) {
-                OsReleaseSIGIO();
+                release_sigio();
                 return 0;
             }
             sigemptyset(&sa.sa_mask);
             sigaddset(&sa.sa_mask, SIGIO);
-            sa.sa_flags = 0;
+            sa.sa_flags = SA_RESTART;
             sa.sa_handler = xf86SIGIO;
             sigaction(SIGIO, &sa, &osa);
             xf86SigIOFuncs[i].fd = fd;
@@ -186,10 +233,8 @@ xf86InstallSIGIOHandler(int fd, void (*f) (int, void *), void *closure)
             xf86SigIOFuncs[i].f = f;
             if (i >= xf86SigIOMax)
                 xf86SigIOMax = i + 1;
-            if (fd >= xf86SigIOMaxFd)
-                xf86SigIOMaxFd = fd + 1;
-            FD_SET(fd, &xf86SigIOMask);
-            OsReleaseSIGIO();
+            xf86SigIOAdd(fd);
+            release_sigio();
             return 1;
         }
         /* Allow overwriting of the closure and callback */
@@ -209,14 +254,12 @@ xf86RemoveSIGIOHandler(int fd)
     struct sigaction osa;
     int i;
     int max;
-    int maxfd;
     int ret;
 
     if (!xf86Info.useSIGIO)
         return 0;
 
     max = 0;
-    maxfd = -1;
     ret = 0;
     for (i = 0; i < MAX_FUNCS; i++) {
         if (xf86SigIOFuncs[i].f) {
@@ -224,13 +267,11 @@ xf86RemoveSIGIOHandler(int fd)
                 xf86SigIOFuncs[i].f = 0;
                 xf86SigIOFuncs[i].fd = 0;
                 xf86SigIOFuncs[i].closure = 0;
-                FD_CLR(fd, &xf86SigIOMask);
+                xf86SigIORemove(fd);
                 ret = 1;
             }
             else {
                 max = i + 1;
-                if (xf86SigIOFuncs[i].fd >= maxfd)
-                    maxfd = xf86SigIOFuncs[i].fd + 1;
             }
         }
     }
@@ -247,7 +288,6 @@ xf86RemoveSIGIOHandler(int fd)
         }
 #endif
         xf86SigIOMax = max;
-        xf86SigIOMaxFd = maxfd;
         if (!max) {
             sigemptyset(&sa.sa_mask);
             sigaddset(&sa.sa_mask, SIGIO);
@@ -258,34 +298,3 @@ xf86RemoveSIGIOHandler(int fd)
     }
     return ret;
 }
-
-int
-xf86BlockSIGIO(void)
-{
-    return OsBlockSIGIO();
-}
-
-void
-xf86UnblockSIGIO(int wasset)
-{
-    OsReleaseSIGIO();
-}
-
-void
-xf86AssertBlockedSIGIO(char *where)
-{
-    sigset_t set, old;
-
-    sigemptyset(&set);
-    sigprocmask(SIG_BLOCK, &set, &old);
-    if (!sigismember(&old, SIGIO))
-        xf86Msg(X_ERROR, "SIGIO not blocked at %s\n", where);
-}
-
-/* XXX This is a quick hack for the benefit of xf86SetSilkenMouse() */
-
-int
-xf86SIGIOSupported(void)
-{
-    return 1;
-}
